﻿using System;
using UnityEngine;

public class InputManager : MonoBehaviour {
    private static InputManager instance;

    public static InputManager Instance
    {
        get
        {
            if (instance == null)
            {
                instance = FindObjectOfType(typeof(InputManager))
                    as InputManager;
            }

            return instance;
        }
    }

    private bool IsButtonDownPrev { get; set; }
    private bool IsButtonDown
    {
        get { return Input.GetKey(KeyCode.Space) || Input.GetKey(KeyCode.Joystick1Button0); }
    }
    
    private MotionState PrevMotionState { get; set; }
    public MotionState CurrentMotionState { get; set; }

    public ButtonStatus ButtonStat
    {
        get
        {
            if (!IsButtonDown)
                return ButtonStatus.Released;
            if (IsButtonDownPrev)
                return ButtonStatus.Holding;
            return ButtonStatus.Pressed;
        }
    }

    public InputStatus ShortButtonStat
    {
        get
        {
            if (!IsButtonDownPrev &&
                IsButtonDown)
                return InputStatus.Entered;
            return InputStatus.None;
        }
    }

    public InputStatus LongButtonStat
    {
        get
        {
            if (!IsButtonDownPrev &&
                IsButtonDown)
                return InputStatus.Entered;
            if (IsButtonDownPrev &&
                IsButtonDown)
                return InputStatus.Continuing;
            if (IsButtonDownPrev &&
                !IsButtonDown)
                return InputStatus.Stopped;
            return InputStatus.None;
        }
    }

    public InputStatus ShortMotionStat
    {
        get
        {
            if (ClapStat        == InputStatus.Entered ||
                JumpStat        == InputStatus.Entered ||
                PushUpLeftStat  == InputStatus.Entered ||
                PushUpRightStat == InputStatus.Entered ||
                PushUpBothStat  == InputStatus.Entered ||
                GuardLeftStat   == InputStatus.Entered ||
                GuardRightStat  == InputStatus.Entered ||
                GuardBothStat   == InputStatus.Entered)
                return InputStatus.Entered;
            return InputStatus.None;
        }
    }

    public InputStatus LongMotionStat
    {
        get
        {
            if (HandUpLeftStat     == InputStatus.Entered    ||
                HandUpLeftStat     == InputStatus.Continuing ||
                HandUpRightStat    == InputStatus.Entered    ||
                HandUpRightStat    == InputStatus.Continuing ||
                HandUpBothStat     == InputStatus.Entered    ||
                HandUpBothStat     == InputStatus.Continuing ||
                HandDownLeftStat   == InputStatus.Entered    ||
                HandDownLeftStat   == InputStatus.Continuing ||
                HandDownRightStat  == InputStatus.Entered    ||
                HandDownRightStat  == InputStatus.Continuing ||
                HandDownBothStat   == InputStatus.Entered    ||
                HandDownBothStat   == InputStatus.Continuing ||
                JesusStat          == InputStatus.Entered    ||
                JesusStat          == InputStatus.Continuing ||
                HeadphoneLeftStat  == InputStatus.Entered    ||
                HeadphoneLeftStat  == InputStatus.Continuing ||
                HeadphoneRightStat == InputStatus.Entered    ||
                HeadphoneRightStat == InputStatus.Continuing ||
                HeadphoneBothStat  == InputStatus.Entered    ||
                HeadphoneBothStat  == InputStatus.Continuing ||
                OnTheTableStat     == InputStatus.Entered    ||
                OnTheTableStat     == InputStatus.Continuing)
                return InputStatus.Entered;
            return InputStatus.None;
        }
    }

    public InputStatus ClapStat
    { get { return ShortMotionToInput("Clap"); } }
    public InputStatus JumpStat
    { get { return ShortMotionToInput("Jump"); } }
    public InputStatus PushUpLeftStat
    { get { return ShortMotionToInput("PushUpLeft"); } }
    public InputStatus PushUpRightStat
    { get { return ShortMotionToInput("PushUpRight"); } }
    public InputStatus PushUpBothStat
    {
        get
        {
            var left = PushUpLeftStat;
            var right = PushUpRightStat;
            return left == right ? left : InputStatus.None;
        }
    }
    public InputStatus GuardLeftStat
    { get { return ShortMotionToInput("GuardLeft"); } }
    public InputStatus GuardRightStat
    { get { return ShortMotionToInput("GuardRight"); } }
    public InputStatus GuardBothStat
    {
        get
        {
            var left = GuardLeftStat;
            var right = GuardRightStat;
            return left == right ? left : InputStatus.None;
        }
    }

    public InputStatus HandUpLeftStat
    { get { return LongMotionToInput("HandUpLeft"); } }
    public InputStatus HandUpRightStat
    { get { return LongMotionToInput("HandUpRight"); } }
    public InputStatus HandUpBothStat
    {
        get
        {
            var left = HandUpLeftStat;
            var right = HandUpRightStat;
            return left == right ? left : InputStatus.None;
        }
    }
    public InputStatus HandDownLeftStat
    { get { return LongMotionToInput("HandDownLeft"); } }
    public InputStatus HandDownRightStat
    { get { return LongMotionToInput("HandDownRight"); } }
    public InputStatus HandDownBothStat
    {
        get
        {
            var left = HandDownLeftStat;
            var right = HandDownRightStat;
            return left == right ? left : InputStatus.None;
        }
    }
    public InputStatus JesusStat
    { get { return LongMotionToInput("Jesus"); } }
    public InputStatus HeadphoneLeftStat
    { get { return LongMotionToInput("HeadphoneLeft"); } }
    public InputStatus HeadphoneRightStat
    { get { return LongMotionToInput("HeadphoneRight"); } }
    public InputStatus HeadphoneBothStat
    {
        get
        {
            var left = HeadphoneLeftStat;
            var right = HeadphoneRightStat;
            return left == right ? left : InputStatus.None;
        }
    }
    public InputStatus OnTheTableStat
    { get { return LongMotionToInput("OnTheTable"); } }

    // Update is called once per frame
    void Update()
    {
        IsButtonDownPrev = IsButtonDown;
        PrevMotionState = CurrentMotionState;
    }

    private InputStatus LongMotionToInput(string name)
    {
        Func<MotionState, MotionStatus> extractor = MotionStatusExtractor.GetExtractor(name);
        if (extractor == null)
            return InputStatus.None;

        MotionStatus prev = extractor(PrevMotionState),
                     curr = extractor(CurrentMotionState);

        if (prev == MotionStatus.None &&
            (curr & MotionStatus.Prepared) != MotionStatus.None)
            return InputStatus.Entered;
        if ((prev & MotionStatus.Done) != MotionStatus.None &&
            curr == MotionStatus.None)
            return InputStatus.Stopped;
        if (prev == curr &&
            (prev & (MotionStatus.Prepared | MotionStatus.Done)) != MotionStatus.None)
            return InputStatus.Continuing;

        return InputStatus.None;
    }

    private InputStatus ShortMotionToInput(string name)
    {
        Func<MotionState, MotionStatus> extractor = MotionStatusExtractor.GetExtractor(name);
        if (extractor == null)
            return InputStatus.None;

        MotionStatus prev = extractor(PrevMotionState),
                     curr = extractor(CurrentMotionState);

        if (prev == MotionStatus.Prepared &&
            curr == MotionStatus.Done)
            return InputStatus.Entered;

        return InputStatus.None;
    }

    private static class MotionStatusExtractor
    {
        private static bool IsContain(MotionState target, MotionState motion)
        {
            return (target & motion) == motion;
        }

        // SHORT NOTES

        private static MotionStatus ExtractClap(MotionState state)
        {
            if (IsContain(state, MotionState.CLAP_PREPARE))
                return MotionStatus.Prepared;
            if (IsContain(state, MotionState.CLAP_DONE))
                return MotionStatus.Prepared;
            return MotionStatus.None;
        }

        private static MotionStatus ExtractJump(MotionState state)
        {
            if (IsContain(state, MotionState.JUMP_PREPARE))
                return MotionStatus.Prepared;
            if (IsContain(state, MotionState.JUMP_DONE))
                return MotionStatus.Prepared;
            return MotionStatus.None;
        }

        private static MotionStatus ExtractPushUpLeft(MotionState state)
        {
            if (IsContain(state, MotionState.HURRAY) &&
                IsContain(state, MotionState.HAND_MOVE_UP_LEFT))
                return MotionStatus.Prepared;
            if (IsContain(state, MotionState.HURRAY) &&
                IsContain(state, MotionState.HAND_MOVE_DOWN_LEFT))
                return MotionStatus.Prepared;
            return MotionStatus.None;
        }

        private static MotionStatus ExtractPushUpRight(MotionState state)
        {
            if (IsContain(state, MotionState.HURRAY) &&
                IsContain(state, MotionState.HAND_MOVE_UP_RIGHT))
                return MotionStatus.Prepared;
            if (IsContain(state, MotionState.HURRAY) &&
                IsContain(state, MotionState.HAND_MOVE_DOWN_RIGHT))
                return MotionStatus.Prepared;
            return MotionStatus.None;
        }

        private static MotionStatus ExtractGuardLeft(MotionState state)
        {
            if (IsContain(state, MotionState.GUARD_BASE_LEFT) &&
                IsContain(state, MotionState.HAND_MOVE_UP_LEFT))
                return MotionStatus.Prepared;
            if (IsContain(state, MotionState.GUARD_BASE_LEFT) &&
                IsContain(state, MotionState.HAND_MOVE_DOWN_LEFT))
                return MotionStatus.Prepared;
            return MotionStatus.None;
        }

        private static MotionStatus ExtractGuardRight(MotionState state)
        {
            if (IsContain(state, MotionState.GUARD_BASE_RIGHT) &&
                IsContain(state, MotionState.HAND_MOVE_UP_RIGHT))
                return MotionStatus.Prepared;
            if (IsContain(state, MotionState.GUARD_BASE_RIGHT) &&
                IsContain(state, MotionState.HAND_MOVE_DOWN_RIGHT))
                return MotionStatus.Prepared;
            return MotionStatus.None;
        }

        // LONG NOTES

        private static MotionStatus ExtractHandUpLeft(MotionState state)
        {
            if (IsContain(state, MotionState.HAND_UP_LEFT))
                return MotionStatus.Prepared |
                       MotionStatus.Done;
            return MotionStatus.None;
        }

        private static MotionStatus ExtractHandUpRight(MotionState state)
        {
            if (IsContain(state, MotionState.HAND_UP_RIGHT))
                return MotionStatus.Prepared |
                       MotionStatus.Done;
            return MotionStatus.None;
        }

        private static MotionStatus ExtractHandDownLeft(MotionState state)
        {
            if (IsContain(state, MotionState.HAND_DOWN_LEFT))
                return MotionStatus.Prepared |
                       MotionStatus.Done;
            return MotionStatus.None;
        }

        private static MotionStatus ExtractHandDownRight(MotionState state)
        {
            if (IsContain(state, MotionState.HAND_DOWN_RIGHT))
                return MotionStatus.Prepared |
                       MotionStatus.Done;
            return MotionStatus.None;
        }

        private static MotionStatus ExtractJesus(MotionState state)
        {
            if (IsContain(state, MotionState.JESUS))
                return MotionStatus.Prepared |
                       MotionStatus.Done;
            return MotionStatus.None;
        }

        private static MotionStatus ExtractHeadphoneLeft(MotionState state)
        {
            if (IsContain(state, MotionState.HEADPHONE_LEFT))
                return MotionStatus.Prepared |
                       MotionStatus.Done;
            return MotionStatus.None;
        }

        private static MotionStatus ExtractHeadphoneRight(MotionState state)
        {
            if (IsContain(state, MotionState.HEADPHONE_RIGHT))
                return MotionStatus.Prepared |
                       MotionStatus.Done;
            return MotionStatus.None;
        }

        private static MotionStatus ExtractOnTheTable(MotionState state)
        {
            if (IsContain(state, MotionState.ON_THE_TABLE))
                return MotionStatus.Prepared |
                       MotionStatus.Done;
            return MotionStatus.None;
        }

        public static Func<MotionState, MotionStatus> GetExtractor(string name)
        {
            if (name.Equals("Clap"))
                return ExtractClap;
            if (name.Equals("Jump"))
                return ExtractJump;
            if (name.Equals("PushUpLeft"))
                return ExtractPushUpLeft;
            if (name.Equals("PushUpRight"))
                return ExtractPushUpRight;
            if (name.Equals("GuardLeft"))
                return ExtractGuardLeft;
            if (name.Equals("GuardRight"))
                return ExtractGuardRight;
            if (name.Equals("HandUpLeft"))
                return ExtractHandUpLeft;
            if (name.Equals("HandUpRight"))
                return ExtractHandUpRight;
            if (name.Equals("HandDownLeft"))
                return ExtractHandDownLeft;
            if (name.Equals("HandDownRight"))
                return ExtractHandDownRight;
            if (name.Equals("Jesus"))
                return ExtractJesus;
            if (name.Equals("HeadphoneLeft"))
                return ExtractHeadphoneLeft;
            if (name.Equals("HeadphoneRight"))
                return ExtractHeadphoneRight;
            if (name.Equals("OnTheTable"))
                return ExtractOnTheTable;
            return null;
        }
    }
}

public enum ButtonStatus
{
    Pressed,
    Released,
    Holding
}

[Flags]
public enum MotionStatus : uint
{
    Prepared = 0x1,
    Done = 0x2,
    None = 0x0
}

public enum InputStatus
{
    Entered,
    Continuing,
    Stopped,
    None
}